from is_core.auth.permissions import BasePermission


class RelatedCoreAllowed(BasePermission):
    """
    Grant permission if core permission with the name grants access
    """

    name = None

    def __init__(self, name=None):
        super().__init__()
        if name:
            self.name = name

    def has_permission(self, name, request, view, obj=None):
        return (
            view.related_core.permission.has_permission(self.name or name, request, view, obj)
            if view.related_core else True
        )



from django.core.exceptions import FieldDoesNotExist


def get_model_field_names(model):
    return [model_field.name for model_field in model._meta.fields] + ['pk']


def get_model_field_by_name(model, field_name):
    try:
        return model._meta.get_field_by_name(field_name)[0]
    except FieldDoesNotExist:
        return None


def get_model_field_value(field_name, instance):
    if '__' in field_name:
        current_field_name, next_field_name = field_name.split('__', 1)
        return get_model_field_value(next_field_name, getattr(instance, current_field_name))
    else:
        try:
            callable_value = getattr(instance, 'get_%s_display' % field_name, None)
            if not callable_value:
                callable_value = getattr(instance, field_name)
            if hasattr(callable_value, '__call__'):
                value = callable_value()
            else:
                value = callable_value
            return value
        except ValueError:
            return ''



from django.views.generic.base import ContextMixin
from django.template.loader import render_to_string

from is_core.auth.views import FieldPermissionViewMixin
from is_core.auth.permissions import AllowAny, FieldsSetPermission

from .permissions import RelatedCoreAllowed


class InlineView(ContextMixin):

    template_name = None

    permission = AllowAny()

    def __init__(self, request, parent_view, parent_instance):
        self.request = request
        self.parent_view = parent_view
        self.parent_instance = parent_instance
        assert self.permission, 'Permissions must be set'

    def can_render(self):
        return self.permission.has_permission('read', self.request, self)

    def render(self, context, title):
        assert self.can_render(), 'Inline view cannot be rendered'

        template = self.template_name
        context.update(self.get_context_data(title=title))
        return render_to_string(template, context.flatten())

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


class RelatedInlineView(FieldPermissionViewMixin, InlineView):

    permission = RelatedCoreAllowed()
    model = None

    def _get_field_permissions(self):
        related_core = self._get_related_core()
        if related_core:
            return self.field_permissions if self.field_permissions is not None else self.related_core.field_permissions
        else:
            return self.field_permissions if self.field_permissions else FieldsSetPermission()

    def __init__(self, request, parent_view, parent_instance):
        super().__init__(request, parent_view, parent_instance)
        self.related_core = self._get_related_core()

    def _get_related_core(self):
        from is_core.site import get_model_core

        return get_model_core(self.model)



from django.forms.formsets import DELETION_FIELD_NAME
from django.utils.translation import ugettext_lazy as _
from django.utils.functional import cached_property

from chamber.utils.forms import formset_has_file_field

from is_core.forms.models import BaseInlineFormSet, smartinlineformset_factory, SmartModelForm
from is_core.generic_views.inlines import RelatedInlineView
from is_core.forms.fields import SmartReadonlyField, EmptyReadonlyField
from is_core.utils import get_readonly_field_data, GetMethodFieldMixin


class InlineFormView(GetMethodFieldMixin, RelatedInlineView):

    form_class = SmartModelForm
    base_inline_formset_class = BaseInlineFormSet

    fields = None
    exclude = ()
    inline_views = None
    field_labels = None

    template_name = None

    fk_name = None
    extra = 0
    can_add = True
    can_delete = True
    is_readonly = False
    max_num = None
    min_num = 0
    readonly_fields = ()
    initial = []
    no_items_text = _('There are no items')
    class_names = ['inline-js']
    add_inline_button_verbose_name = None

    def __init__(self, request, parent_view, parent_instance):
        super().__init__(request, parent_view, parent_instance)
        self.core = parent_view.core
        self.parent_instance = parent_instance
        self.parent_model = self.parent_instance.__class__
        self.readonly = self._is_readonly()

    @cached_property
    def formset(self):
        return self.get_formset()

    def _get_field_labels(self):
        return self.field_labels

    def _is_readonly(self):
        return self.is_readonly or self.parent_view.is_readonly()

    def can_form_delete(self, form):
        return (
            not self.is_form_readonly(form)
            and self.permission.has_permission('delete', self.request, self, obj=form.instance)
        )

    def is_form_readonly(self, form):
        return self.readonly and self.permission.has_permission('update', self.request, self, obj=form.instance)

    def get_context_data(self, **kwargs):
        formset = self.formset
        context_data = super().get_context_data(**kwargs)
        context_data.update({
            'formset': formset,
            'fields': self.get_fromset_fields(formset),
            'name': self.get_name(),
            'button_value': self.get_button_value(),
            'class_names': self.get_class_names(formset, **kwargs),
            'no_items_text': self.no_items_text
        })
        return context_data

    def get_class_names(self, formset, **kwargs):
        class_names = self.class_names + [self.get_name().lower()]
        if formset.can_add:
            class_names.append('can-add')
        if formset.can_delete:
            class_names.append('can-delete')

        if kwargs.get('title'):
            class_names.append('with-title')
        else:
            class_names.append('without-title')

        return class_names

    def get_exclude(self):
        return self.exclude

    def generate_fields(self):
        fields = self.get_fields()
        if fields is None:
            fields = (
                list(self.get_form_class().base_fields.keys())
                + list(self.get_formset_factory().form.base_fields.keys())
            )
        return [field for field in fields if field not in self._get_disallowed_fields_from_permissions()]

    def get_fields(self):
        return self.fields

    def get_extra(self):
        return self.extra + len(self.get_initial())

    def get_initial(self):
        return self.initial[:]

    def get_can_delete(self):
        return (
            (self.get_can_add() or (self.can_delete and self.permission.has_permission('delete', self.request, self)))
            and not self.readonly
        )

    def get_can_add(self):
        return self.can_add and not self.readonly and self.permission.has_permission('create', self.request, self)

    def get_readonly_fields(self):
        return self.readonly_fields

    def generate_readonly_fields(self):
        return list(self.get_readonly_fields()) + list(self._get_readonly_fields_from_permissions())

    def get_prefix(self):
        return '-'.join((self.parent_view.get_prefix(), 'inline', self.__class__.__name__)).lower()

    def get_fromset_fields(self, formset):
        fields = list(self.generate_fields() or formset.form.base_fields.keys())
        if formset.can_delete:
            fields.append(DELETION_FIELD_NAME)
        return fields

    def formfield_for_dbfield(self, db_field, **kwargs):
        return db_field.formfield(**kwargs)

    def formfield_for_readonlyfield(self, name, **kwargs):
        def _get_readonly_field_data(instance):
            return get_readonly_field_data(instance, name, self.request, view=self)
        return SmartReadonlyField(_get_readonly_field_data)

    def get_form_class(self):
        return self.form_class

    def get_max_num(self):
        return self.max_num

    def get_min_num(self):
        return self.min_num

    def get_formset_factory(self, fields=None, readonly_fields=()):
        return smartinlineformset_factory(
            self.parent_model, self.model, self.request, form=self.get_form_class(), fk_name=self.fk_name,
            extra=self.get_extra(), formset=self.base_inline_formset_class, exclude=self.get_exclude(),
            fields=fields, min_num=self.get_min_num(), max_num=self.get_max_num(), readonly_fields=readonly_fields,
            readonly=self._is_readonly(), formreadonlyfield_callback=self.formfield_for_readonlyfield,
            formfield_callback=self.formfield_for_dbfield, labels=self._get_field_labels(),
            can_delete=self.get_can_delete()
        )

    def get_queryset(self):
        return self.model.objects.all()

    def get_formset(self):
        fields = self.generate_fields()
        readonly_fields = self.generate_readonly_fields()

        if self.request.POST:
            formset = self.get_formset_factory(fields, readonly_fields)(data=self.request.POST,
                                                                        files=self.request.FILES,
                                                                        instance=self.parent_instance,
                                                                        queryset=self.get_queryset(),
                                                                        prefix=self.get_prefix())
        else:
            formset = self.get_formset_factory(fields, readonly_fields)(instance=self.parent_instance,
                                                                        queryset=self.get_queryset(),
                                                                        initial=self.get_initial(),
                                                                        prefix=self.get_prefix())

        formset.can_add = self.get_can_add()
        for form in formset.all_forms():
            form.class_names = self.form_class_names(form)
            form._is_readonly = self.is_form_readonly(form)
            if not self.readonly and form._is_readonly:
                if formset.can_delete:
                    form.readonly_fields = set(form.fields.keys()) - {'id', DELETION_FIELD_NAME}
                else:
                    form.readonly_fields = set(form.fields.keys()) - {'id'}

            if formset.can_delete and form.instance.pk and not self.can_form_delete(form):
                form.fields[DELETION_FIELD_NAME] = EmptyReadonlyField(
                    required=form.fields[DELETION_FIELD_NAME].required,
                    label=form.fields[DELETION_FIELD_NAME].label
                )
                form.readonly_fields |= {DELETION_FIELD_NAME}

            self.init_form(form)

        for i in range(self.get_min_num()):
            formset.forms[i].empty_permitted = False
        return formset

    def form_class_names(self, form):
        if not form.instance.pk:
            return ['empty']
        return []

    def init_form(self, form):
        self.form_fields(form)

    def form_fields(self, form):
        for field_name, field in form.fields.items():
            self.form_field(form, field_name, field)

    def form_field(self, form, field_name, form_field):
        placeholder = self.model._ui_meta.placeholders.get('field_name', None)
        if placeholder:
            form_field.widget.placeholder = self.model._ui_meta.placeholders.get('field_name', None)
        return form_field

    def get_name(self):
        return self.model.__name__

    def get_button_value(self):
        return self.add_inline_button_verbose_name or self.model._ui_meta.add_inline_button_verbose_name % {
            'verbose_name': self.model._meta.verbose_name,
            'verbose_name_plural': self.model._meta.verbose_name_plural
        }

    def form_valid(self, request):
        formset = self.formset
        instances = formset.save(commit=False)
        for obj in instances:
            change = obj.pk is not None
            self.save_obj(obj, change)
        for obj in formset.deleted_objects:
            self.delete_obj(obj)
        formset.save_m2m()

    def get_has_file_field(self):
        return formset_has_file_field(self.formset.form)

    def pre_save_obj(self, obj, change):
        pass

    def post_save_obj(self, obj, change):
        pass

    def save_obj(self, obj, change):
        self.pre_save_obj(obj, change)
        obj.save()
        self.post_save_obj(obj, change)

    def pre_delete_obj(self, obj):
        pass

    def post_delete_obj(self, obj):
        pass

    def delete_obj(self, obj):
        self.pre_delete_obj(obj)
        obj.delete()
        self.post_delete_obj(obj)

    def is_valid(self):
        return self.formset.is_valid()

    def has_changed(self):
        return self.formset.has_changed()


class TabularInlineFormView(InlineFormView):

    template_name = 'is_core/forms/tabular_inline_formset.html'


class StackedInlineFormView(InlineFormView):

    template_name = 'is_core/forms/stacked_inline_formset.html'


class ResponsiveInlineFormView(InlineFormView):

    template_name = 'is_core/forms/responsive_inline_formset.html'



import re

from urllib.parse import urlunparse, urlparse

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.http import QueryDict
from django.http.response import Http404
from django.views.generic.base import TemplateView, View
from django.utils.translation import ugettext_lazy as _
from django.conf import settings
from django.urls import reverse

from block_snippets.views import JSONSnippetTemplateResponseMixin

from is_core.auth.permissions import PermissionsSet, CoreReadAllowed, CoreAllowed, DEFAULT_PERMISSION
from is_core.menu import LinkMenuItem
from is_core.exceptions import (
    HTTPForbiddenResponseException, HTTPUnauthorizedResponseException, HTTPRedirectResponseException
)
from is_core.generic_views.exceptions import GenericViewException


def redirect_to_login(next, redirect_field_name=REDIRECT_FIELD_NAME):
    """
    Redirects the user to the login page, passing the given 'next' page
    """
    resolved_url = reverse('IS:login')

    login_url_parts = list(urlparse(resolved_url))
    if redirect_field_name:
        querystring = QueryDict(login_url_parts[4], mutable=True)
        querystring[redirect_field_name] = next
        login_url_parts[4] = querystring.urlencode(safe='/')

    raise HTTPRedirectResponseException(urlunparse(login_url_parts))


class PermissionsViewMixin:

    permission = None

    auto_login_redirect = True

    def __init__(self):
        super().__init__()
        assert self.permission, 'Permissions must be set'

    def dispatch(self, request, *args, **kwargs):
        rm = request.method.lower()
        if rm in self.http_method_names and hasattr(self, rm):
            self._check_permission(rm)
            handler = getattr(self, rm)
        else:
            handler = self.http_method_not_allowed
        return handler(request, *args, **kwargs)

    def has_permission(self, name, obj=None):
        # For core obj view mixin is object required to check permissions. If no object is found False is returned.
        try:
            return self._has_permission(name, obj)
        except Http404:
            return False

    def _has_permission(self, name, obj=None):
        return self.permission.has_permission(name, self.request, self, obj)

    def _check_permission(self, name, obj=None):
        """
        If customer is not authorized he should not get information that object is exists.
        Therefore 403 is returned if object was not found or is redirected to the login page.
        If custmer is authorized and object was not found is returned 404.
        If object was found and user is not authorized is returned 403 or redirect to login page.
        If object was found and user is authorized is returned 403 or 200 according of result of _has_permission method.
        """
        def redirect_or_exception(ex):
            if not self.request.user or not self.request.user.is_authenticated:
                if self.auto_login_redirect:
                    redirect_to_login(self.request.get_full_path())
                else:
                    raise HTTPUnauthorizedResponseException
            else:
                raise ex

        try:
            if not self._has_permission(name, obj):
                redirect_or_exception(HTTPForbiddenResponseException)
        except Http404 as ex:
            redirect_or_exception(ex)


class DefaultViewMixin(PermissionsViewMixin, JSONSnippetTemplateResponseMixin):

    site_name = None
    menu_groups = None
    allowed_snippets = ('content',)
    view_name = None
    add_current_to_breadcrumbs = True
    kwargs = {}
    args = {}
    title = None
    page_title = None

    def __init__(self):
        super(DefaultViewMixin, self).__init__()

    def get_title(self):
        return self.title

    def get_page_title(self):
        return self.page_title or self.get_title()

    def get_context_data(self, **kwargs):
        context_data = super(DefaultViewMixin, self).get_context_data(**kwargs)
        extra_context_data = {
            'site_name': self.site_name,
            'active_menu_groups': self.menu_groups,
            'title': self.get_title(),
            'page_title': self.get_page_title(),
            'view_name': self.view_name,
            'bread_crumbs_menu_items': self.bread_crumbs_menu_items(),
        }
        extra_context_data.update(context_data)
        return extra_context_data

    def bread_crumbs_menu_items(self):
        if self.add_current_to_breadcrumbs:
            return [LinkMenuItem(self.get_title(), self.request.path, active=True)]
        return []


class DefaultCoreViewMixin(DefaultViewMixin):

    core = None
    view_type = None
    title = None
    permission = PermissionsSet(
        get=CoreReadAllowed(),
        **{
            DEFAULT_PERMISSION: CoreAllowed(),
        }
    )

    @classmethod
    def __init_core__(cls, core, pattern):
        cls.core = core
        cls.pattern = pattern
        cls.site_name = core.site_name
        cls.menu_groups = core.get_menu_groups()

    def dispatch(self, request, *args, **kwargs):
        self.core.init_ui_request(request)
        return super(DefaultCoreViewMixin, self).dispatch(request, *args, **kwargs)

    def get_title(self):
        return self.title

    @property
    def view_name(self):
        return '%s-%s' % (self.view_type, '-'.join(self.menu_groups))

    def get_context_data(self, **kwargs):
        context_data = super(DefaultCoreViewMixin, self).get_context_data(**kwargs)
        extra_context_data = {
            'core_permission': self.core.permission,
            'permission': self.permission,
            'view': self
        }
        extra_context_data.update(context_data)
        return extra_context_data


class DefaultModelCoreViewMixin(DefaultCoreViewMixin):

    model = None

    @property
    def verbose_name(self):
        return self.core.verbose_name

    @property
    def verbose_name_plural(self):
        return self.core.verbose_name_plural

    @classmethod
    def __init_core__(cls, core, pattern):
        super(DefaultModelCoreViewMixin, cls).__init_core__(core, pattern)
        cls.model = cls.model or core.model

    def get_context_data(self, **kwargs):
        context_data = super(DefaultModelCoreViewMixin, self).get_context_data(**kwargs)
        extra_context_data = {
            'core': self.core
        }
        extra_context_data.update(context_data)
        return extra_context_data

    def get_title(self):
        return self.title or self.verbose_name


class HomeView(DefaultCoreViewMixin, TemplateView):

    template_name = 'is_core/home.html'
    view_name = 'home'

    def get_title(self):
        return _('Home')



from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.forms.models import ModelForm
from django.contrib.contenttypes.forms import BaseGenericInlineFormSet as OriginBaseGenericInlineFormSet

from is_core.forms.models import smartmodelformset_factory
from is_core.forms.formsets import BaseFormSetMixin


class BaseGenericInlineFormSet(BaseFormSetMixin, OriginBaseGenericInlineFormSet):
    pass


def smart_generic_inlineformset_factory(model, request, form=ModelForm, formset=BaseGenericInlineFormSet,
                                        ct_field='content_type', fk_field='object_id', fields=None, exclude=None,
                                        extra=3, can_order=False, can_delete=True, min_num=None, max_num=None,
                                        formfield_callback=None, widgets=None, validate_min=False, validate_max=False,
                                        localized_fields=None, labels=None, help_texts=None, error_messages=None,
                                        formreadonlyfield_callback=None, readonly_fields=None, for_concrete_model=True,
                                        readonly=False):
    """
    Returns a ``GenericInlineFormSet`` for the given kwargs.

    You must provide ``ct_field`` and ``fk_field`` if they are different from
    the defaults ``content_type`` and ``object_id`` respectively.
    """
    opts = model._meta
    # if there is no field called `ct_field` let the exception propagate
    ct_field = opts.get_field(ct_field)
    if not isinstance(ct_field, models.ForeignKey) or ct_field.related_model != ContentType:
        raise Exception("fk_name '%s' is not a ForeignKey to ContentType" % ct_field)
    fk_field = opts.get_field(fk_field)  # let the exception propagate
    if exclude is not None:
        exclude = list(exclude)
        exclude.extend([ct_field.name, fk_field.name])
    else:
        exclude = [ct_field.name, fk_field.name]

    kwargs = {
        'form': form,
        'formfield_callback': formfield_callback,
        'formset': formset,
        'extra': extra,
        'can_delete': can_delete,
        'can_order': can_order,
        'fields': fields,
        'exclude': exclude,
        'max_num': max_num,
        'min_num': min_num,
        'widgets': widgets,
        'validate_min': validate_min,
        'validate_max': validate_max,
        'localized_fields': localized_fields,
        'formreadonlyfield_callback': formreadonlyfield_callback,
        'readonly_fields': readonly_fields,
        'readonly': readonly,
        'labels': labels,
        'help_texts': help_texts,
        'error_messages': error_messages,
    }
    FormSet = smartmodelformset_factory(model, request, **kwargs)
    FormSet.ct_field = ct_field
    FormSet.ct_fk_field = fk_field
    FormSet.for_concrete_model = for_concrete_model
    return FormSet
